# Value Function Iteration to solve the stochastic  
# Single-State Optimal Growth model 
#
# Original AMPL coding: June 30, 2006

# important parameters
param beta := 0.8;    # discount factor
param gamma := 2;
param eta := 1;
param n := 11;        # number of Chebyshev nodes
param deg := 10;      # degree of Chebyshev polynomial
param kmin := 0.5;    # minimum state value
param kmax := 1.5;    # maximum state value

# Define some constants
param pi  := 3.14159265358979;
param eps := 1e-6;

# Define uncertainty and its values and probablity transition matrix
param thetaN := 2;
set   Stheta := 1..thetaN;
param thetas {i in Stheta};
param PM {i in Stheta, j in Stheta};
param PDF {i in Stheta} := 1/thetaN; # uniform probability density function
let thetas[1] := 0.9;
let thetas[2] := 1.1;
let PM[1,1] := 0.75; 
let PM[1,2] := 0.25; 
let PM[2,1] := 0.25; 
let PM[2,2] := 0.75; 

# run time
param runTime default 0;

# Define parameters values
set   SN := 1..n;       # index set of states
set   Sdeg := 1..(deg+1);
param alpha := 0.25;
param A := (1-beta)/(alpha*beta);
param B := (1-alpha)*A^(1-gamma);

# Define Chebyshev nodes
param z{i in SN} := - cos((2*i-1)/(2*n)*pi); # the Chebyshev interpolation nodes on [-1,1]
param ext := -(z[1]+1)*(kmax-kmin)/2/z[1];
param extkmin := kmin - ext;
param extkmax := kmax + ext;
param dz := 2/(extkmax-extkmin); # dz/dx 
param KNodes {i in SN} := extkmin + (1+z[i])/dz;  # set of Chebyshev state variables

# Define variables
var coefs {i in Sdeg, j in Stheta}; # coefficients
var kPlus {i in SN, j in Stheta}; # tomorrow capital
var L {i in SN, j in Stheta} >= eps; # labor supply
var c {i in SN, j in Stheta} >= eps; # consumptions

# model for Bellman
maximize TV : sum {k1 in Stheta} (sum {i in SN} ( (c[i,k1]^(1-gamma)/(1-gamma) - B*L[i,k1]^(1+eta)/(1+eta)) + 
                  beta*(sum {k2 in Stheta} (PM[k2,k1] * (sum {j in Sdeg} 
                  (coefs[j,k2]*cos((j-1)*acos(dz*(kPlus[i,k1]-extkmin)-1))) ))) ));
subject to  Balance {i in SN, k1 in Stheta} : 
            kPlus[i,k1] = KNodes[i] + thetas[k1]*A*(KNodes[i]^alpha)*(L[i,k1]^(1-alpha)) - c[i,k1];
subject to  kL {i in SN, k1 in Stheta} : kPlus[i,k1] >= kmin;
subject to  kU {i in SN, k1 in Stheta} : kPlus[i,k1] <= kmax;

# define Chebyshev basis polynomial and their derivatives values at KNodes 
# by use of the recursive formula
param T {i in SN, j in Sdeg} default 0; # Chebyshev polynomial values at KNodes
param dT {i in SN, j in Sdeg} default 0; # Chebyshev polynomial first derivative values at KNodes
param ddT {i in SN, j in Sdeg} default 0; # Chebyshev polynomial second derivative values at KNodes
for {i in SN} {
  let T[i,1] := 1;
  let T[i,2] := z[i]; 
  let dT[i,1] := 0;
  let dT[i,2] := 1 * dz; 
  let ddT[i,1] := 0;
  let ddT[i,2] := 0;
  for {j in 2..deg} {
    let T[i,j+1] := 2*z[i]*T[i,j] - T[i,j-1];
    let dT[i,j+1] := 2*T[i,j]*dz + 2*z[i]*dT[i,j] - dT[i,j-1];
    let ddT[i,j+1] := 4*dT[i,j]*dz + 2*z[i]*ddT[i,j] - ddT[i,j-1];
  }
}

param V {i in SN, j in Stheta} default 0; # current value function's values at KNodes

######################################

problem Bellman: TV, kPlus, L, c, Balance, kL, kU;

######################################

param MAX_count default 1000;
param counter, default 0;
param err default 1.0;
param err2 default 1.0;

# Define Uniform state values
param  un := 31;         # number of states
set    SUN := 1..un;     # index set of states
param  UKNodes {i in SUN} := kmin + (kmax-kmin)/(un-1)*(i-1);  # set of states 

param Va_plus {k in SUN, j in Stheta}; # tomorrow value function's values at UKNodes
param Va {k in SUN, j in Stheta}; # current value function's values at UKNodes

# define Chebyshev basis polynomial values at UKNodes 
# by use of the recursive formula
param Uz {i in SUN} := dz*(UKNodes[i]-extkmin) - 1;  # normalize UKNodes
param UT {i in SUN, j in Sdeg} default 0; # Chebyshev polynomial values at UKNodes
for {i in SUN} {
  let UT[i,1] := 1;
  let UT[i,2] := Uz[i]; 
  for {j in 2..deg} {
    let UT[i,j+1] := 2*Uz[i]*UT[i,j] - UT[i,j-1];
  }
}

let {i in Sdeg, k1 in Stheta} coefs[i,k1] := 0;
let {k in SUN, k1 in Stheta} Va[k,k1] := sum {j in Sdeg} coefs[j,k1]*UT[k,j];

repeat { 
  option minos_options 'Superbasics_limit=500'; 
  solve Bellman;
  let runTime := runTime + _solve_time;
  let {i in SN, k1 in Stheta} V[i,k1] := (c[i,k1]^(1-gamma)/(1-gamma) - B*L[i,k1]^(1+eta)/(1+eta)) + 
      beta*(sum {k2 in Stheta} (PM[k2,k1] * (sum {j in Sdeg} 
      (coefs[j,k2]*cos((j-1)*acos(dz*(kPlus[i,k1]-extkmin)-1))) ))); 

  # compute Chebyshev coeffcients from V
  let {k1 in Stheta} coefs[1,k1] := (sum {i in SN} (T[i,1]*V[i,k1])) / n;
  let {j in 2..(deg+1), k1 in Stheta} coefs[j,k1] := 2*(sum {i in SN} (T[i,j]*V[i,k1])) / n;

  let {k in SUN, k1 in Stheta} Va_plus[k,k1] := Va[k,k1];
  let {k in SUN, k1 in Stheta} Va[k,k1] := sum {j in Sdeg} coefs[j,k1]*UT[k,j];
  let err := max {k in SUN, k1 in Stheta} abs(Va_plus[k,k1]-Va[k,k1]);
  let err2 := sqrt(sum {k in SUN, k1 in Stheta} (Va_plus[k,k1]-Va[k,k1])^2/(un*thetaN));
  if (counter mod 10 = 0) then { 
    display counter, err, err2;
  }
  let counter := counter + 1;
} while ( err > eps) && (counter < MAX_count);

let runTime := runTime + _ampl_time;

display Va;

display counter, runTime, err, err2, coefs;

# compute slopes and curvatures at nodes
param slope {i in SN, k1 in Stheta};
let {i in SN, k1 in Stheta} slope[i,k1] := sum {j in 2..(deg+1)} (coefs[j,k1]*dT[i,j]);
param curvature {i in SN, k1 in Stheta};
let {i in SN, k1 in Stheta} curvature[i,k1] := sum {j in 3..(deg+1)} (coefs[j,k1]*ddT[i,j]);

display {k1 in Stheta, i in SN} (thetas[k1], KNodes[i], kPlus[i,k1], L[i,k1], c[i,k1], V[i,k1], 
  slope[i,k1], curvature[i,k1]); 

########## END
